import configparser
import tensorflow as tf
from mppi import MPPI
from model import Heat2D
import numpy as np
from simulate import Simulate

np_dtype = np.float32
tf_dtype = tf.float32

seed = np.random.seed(1)
tfseed = tf.random.set_random_seed(1)
config = configparser.RawConfigParser()
config.read('2Dheat_sim.cfg')

dt = config.getfloat('sec1', 'dt') #time increment
Tf = config.getfloat('sec1', 'Tf') #final time
a = config.getfloat('sec1', 'a') #length of heat line
J = config.getint('sec1', 'J') #number of noise expansion terms
iters = config.getint('sec1', 'iters') #number of MPC iterations per time step
rollouts = config.getint('sec1', 'rollouts') #number of rollouts
rho = config.getfloat('sec1', 'rho') #acts as "temperature", squared inverse of variance
learnrate = config.getfloat('sec1', 'learnrate') # learning rate for updates
epsilon = config.getfloat('sec1', 'epsilon') #heat pde constant (scaling of h_xx)
l2_scale = config.getfloat('sec1', 'l2_scale') #regularization scaling
#boundary_conds = config.getstring('sec1', 'boundary_conds')
desired_value = np.float32(config.getfloat('sec1', 'desired_value')) #goal temperature for MPC
Q = config.getfloat('sec1', 'Q') #a single value
load_check = config.getboolean('sec1', 'load_network_check')
use_bias = config.getboolean('sec1', 'use_bias')
relu_activation = config.getboolean('sec1', 'relu_activation')

mpc_steps = int(np.ceil(Tf/dt)) # total number of MPC timesteps

#num_contour_levels = config.getint('sec1', 'num_contour_levels') #number of levels to plot
#vmin = config.getfloat('sec1', 'vmin') #min contour level to plot
#vmax = config.getfloat('sec1', 'vmax') #max contour level to plot


# Get/compute actuator data
actuator_x_locs = np.asarray(config.get('sec2', 'actuator_x_locs').split(','), dtype=np.float32) * np.float32(a) #actuator locations (x positions)
actuator_y_locs = np.asarray(config.get('sec2', 'actuator_y_locs').split(','), dtype=np.float32) * np.float32(a) #actuator locations (y positions)
N = len(actuator_x_locs) #number of actuators

actuator_variance = np.float32(config.getfloat('sec2','actuator_variance') * a)
actuator_variance = np.square(actuator_variance) * np.ones(N)
actuator_variance_xx, actuator_variance_yy = actuator_variance, actuator_variance

num_filters = np.asarray(config.get('sec2', 'num_filters').split(','), dtype=np.int32)
kernel_size = np.asarray(config.get('sec2', 'kernel_size').split(','), dtype=np.int32)
stride = np.asarray(config.get('sec2', 'stride').split(','), dtype=np.int32)

hidden_layers = np.asarray(config.get('sec2', 'hidden_layers').split(','), dtype=np.int32) #list of hidden nodes in each layer 



class Params(object):
	def __init__(self):
		self.relu_activation = relu_activation		
		if relu_activation:
			self.activation = tf.nn.relu 
		else:
			self.activation = tf.nn.tanh
		self.use_bias = use_bias
		self.l2_scale = l2_scale
		self.dt = dt
		self.a = a
		self.bj = np.sqrt(2.0 * self.dt/self.a)
		self.iterations = iters #
		self.learnrate = learnrate
		self.J = J 
		self.avg_spacing = a/J
		self.epsilon = epsilon
		self.rho = rho #tuning parameter
		self.N = N
		self.mpc_steps = mpc_steps
		self.rollouts = rollouts #tuning parameter
		self.actuator_x_locs = actuator_x_locs
		self.actuator_y_locs = actuator_y_locs
		self.actuator_variance_xx = actuator_variance_xx
		self.actuator_variance_yy = actuator_variance_yy

		self.hidden_layers = hidden_layers
		self.Q = Q #tuning parameter
		self.desired_state = np.zeros((1, J+1), dtype=np_dtype)
		#self.h_0 = tf.multiply(tf.ones((rollouts, J+1), dtype=np_dtype), tf.random.normal((1, J+1), dtype=tf_dtype))
		#print("Initial state: ",self.h_0)
		self.h_0 = tf.zeros((rollouts, J+1, J+1), dtype=tf_dtype)

		# self.num_layers = len(hidden_layers)-1

		self.range_start_x = np.int32([np.round(0.48*(J-1)), np.round(0.18*(J-1)), np.round(0.78*(J-1)), np.round(0.48*(J-1)), np.round(0.48*(J-1))])
		self.range_end_x = np.int32([np.round(0.52*(J-1)), np.round(0.22*(J-1)), np.round(0.82*(J-1)), np.round(0.52*(J-1)), np.round(0.52*(J-1))])
		self.range_start_y = np.int32([np.round(0.48*(J-1)), np.round(0.48*(J-1)), np.round(0.48*(J-1)), np.round(0.18*(J-1)), np.round(0.78*(J-1))])
		self.range_end_y = np.int32([np.round(0.52*(J-1)), np.round(0.52*(J-1)), np.round(0.52*(J-1)), np.round(0.22*(J-1)), np.round(0.82*(J-1))])

		self.desired_state = np.zeros((J-1,J-1), dtype=np_dtype)
		self.desired_state[ self.range_start_x[0]: self.range_end_x[0]+1, self.range_start_y[0]: self.range_end_y[0]+1 ] = 0.5*desired_value #middle region
		for i in range(1, len(self.range_start_x)):
			self.desired_state[ self.range_start_x[i]: self.range_end_x[i]+1, self.range_start_y[i]: self.range_end_y[i]+1 ] = desired_value #other regions
		print("Desired state: \n", self.desired_state)

		range_x_1 = np.arange( np.int32( np.round(0.48*(J-1)) ),  np.int32( np.round(0.52*(J-1))+1 ), 1) 
		range_x_2 = np.arange( np.int32( np.round(0.18*(J-1)) ),  np.int32( np.round(0.22*(J-1))+1 ), 1) 
		range_x_3 = np.arange( np.int32( np.round(0.78*(J-1)) ),  np.int32( np.round(0.82*(J-1))+1 ), 1) 
		range_x_4 = np.arange( np.int32( np.round(0.48*(J-1)) ),  np.int32( np.round(0.52*(J-1))+1 ), 1) 
		range_x_5 = np.arange( np.int32( np.round(0.48*(J-1)) ),  np.int32( np.round(0.52*(J-1))+1 ), 1) 

		range_y_1 = np.arange( np.int32( np.round(0.48*(J-1)) ),  np.int32( np.round(0.52*(J-1))+1 ), 1) 
		range_y_2 = np.arange( np.int32( np.round(0.48*(J-1)) ),  np.int32( np.round(0.52*(J-1))+1 ), 1) 
		range_y_3 = np.arange( np.int32( np.round(0.48*(J-1)) ),  np.int32( np.round(0.52*(J-1))+1 ), 1) 
		range_y_4 = np.arange( np.int32( np.round(0.18*(J-1)) ),  np.int32( np.round(0.22*(J-1))+1 ), 1) 
		range_y_5 = np.arange( np.int32( np.round(0.78*(J-1)) ),  np.int32( np.round(0.82*(J-1))+1 ), 1) 

		desired_state = []
		desired_state.append( np.ones((len(range_x_1), len(range_y_1)), dtype=np_dtype) * 0.5 * desired_value )
		desired_state.append( np.ones((len(range_x_2), len(range_y_2)), dtype=np_dtype) * desired_value )
		desired_state.append( np.ones((len(range_x_3), len(range_y_3)), dtype=np_dtype) * desired_value )
		desired_state.append( np.ones((len(range_x_4), len(range_y_4)), dtype=np_dtype) * desired_value )
		desired_state.append( np.ones((len(range_x_5), len(range_y_5)), dtype=np_dtype) * desired_value )
		self.num_desired_regions = len(desired_state)
		self.desired_region_length = len(range_x_1)
		self.desired_state_mat = tf.constant(np.concatenate(desired_state, axis=0), dtype=tf_dtype)

		self.desired_value = desired_value

		# self.num_weights = (J+1)*hidden_layers[0] + hidden_layers[0]
		# for j in range(self.num_layers):
		# 	self.num_weights += hidden_layers[j]*hidden_layers[j+1] + hidden_layers[j+1]
		# self.num_weights += hidden_layers[-1]*N + N

		self.num_filters = num_filters
		self.kernel_size = kernel_size
		self.stride = stride

		self.random_seed = seed
		self.tf_random_seed = tfseed

		self.load_check = load_check

params = Params()
heat2d = Heat2D(params)
sim = Simulate(params, heat2d)
sim._simulate()



		